/**************************************************************************
 *                                                                        *
 *               Copyright (C) 1997, Dynamix, Inc.                        *
 *                                                                        *
 *  These coded instructions, statements, and computer programs  contain  *
 *  unpublished  proprietary  information of Dynamix, Inc.,  and are      *
 *  protected by Federal copyright  law.  They  may not be disclosed to   *
 *  third parties  or copied or duplicated in any form, in whole or       *
 *  in part, without the prior written consent of Dynamix, Inc.           *
 *                                                                        *
 *************************************************************************/


#include <ultra64.h>

#include "main.h"
#include "stdio.h"
#include "static.h"
#include "model.h"
#include "cont.h"
#include "motor.h"

#include "debug.c"
#include "boot.h"

extern OSMesgQueue      n_siMessageQ;

extern unsigned short blue1[];
extern unsigned short blue2[];
extern unsigned short blue3[];

/*
 * calls from scroller.c
 *
 */

extern void scrollerInit(void);
extern void drawScroller(Gfx **glistp);
extern void doScroll(f32 moveCnt);
extern s16 addLetterToScroller(void);
extern s16 scrollSpacer;


/*
 * calls from logo.c
 *
 */

extern void drawLogo(Gfx **glistp,f32 moveCnt);

extern void drawHelp(Gfx **glistp);

/*
 * external variables and definitions for rom mapping addresses
 * used for DMA'ing data from rom to ram
 *
 */

extern char _static_seg_org[],_static_seg_orgend[];
#define _staticSegmentRomStart (_static_seg_org)
#define _staticSegmentRomEnd (_static_seg_orgend)

extern char _rom_seg_org[];
#define _DataSegmentRomStart (_rom_seg_org)

extern char _rom_seg_data_orgend[];
#define _DataSegmentRomEnd (_rom_seg_data_orgend)

extern char _cfb_obj[];
#define _cfbSegmentStart (_cfb_obj)

extern char _zbuf_obj[];
#define _zbufSegmentStart (_zbuf_obj)

extern char _code_bss_objend[];
#define _codeSegmentEnd (_code_bss_objend)

/* Give static segment a place in RAM */
char   *staticSegment = _codeSegmentEnd;


/*
 * external variables for thread structures and messages queues
 * all created and defined in BOOT.C
 *
 */

extern OSMesgQueue      retraceMessageQ;
extern OSMesg           retraceMessageBuf;

extern OSMesgQueue      dmaMessageQ;
extern OSMesg           dmaMessageBuf[];
extern OSIoMesg         dmaIOMessageBuf;

extern OSMesgQueue      rdpMessageQ;
extern OSMesg           rdpMessageBuf;

extern OSMesgQueue      rspMessageQ;
extern OSMesg           rspMessageBuf;

extern OSMesg           dummyMessage;

extern OSContStatus     statusdata[MAXCONTROLLERS];
extern OSContPad        controllerdata[MAXCONTROLLERS];

extern int ostype;

OSPfs                   pfs[MAXCONTROLLERS];

/* printf to screen */

FB              fb;

/* deas stuff */
s32 max_files;
s32 files_used;
u8 names[20*PFS_FILE_NAME_LEN];
u8 name[20*PFS_FILE_NAME_LEN];
u8 ext[20*PFS_FILE_EXT_LEN];
unsigned int fsize[20];
u16 fgc[20];
u32 fcc[20];
int times;
OSPfsState filestate;
u8 filebuffer[32768];
u8 filebuffer2[32768];
s32 errv;
static int anz=0;
static int xxx=0;
int inhelp=0;

uSprite MySprite[5];

/*
 * Double-buffered data structures
 *
 */

dynamic_stuff   dynamic[2];
OSTask          task[2];

/*
 * Stack for SP task
 *
 */

u64 dram_stack[SP_DRAM_STACK_SIZE64];	/* used for matrix stack */

/*
 * Display lists for the 3D objects
 * Can't wait till we have a GL2DL.
 * Well It's not final, but it's a start
 */

int lastent;


/*
 * Draw all sprite objects; logo and scroller
 *
 */

static void drawSprites(Gfx **glistp,f32 moveCnt)
{

  Gfx *gxp;
  gxp = *glistp;

  /* set RM */

  gDPPipeSync(gxp++);
  gDPSetCycleType(gxp++, G_CYC_1CYCLE);
  gDPSetScissor(gxp++, G_SC_NON_INTERLACE, 0, 0, SCREEN_WD, SCREEN_HT);

  spInit(&gxp);
//  drawHelp(&gxp);
  
//  drawLogo(&gxp,moveCnt);

/*
  drawScroller(&gxp);
  if (--scrollSpacer < 0) {
     scrollSpacer = addLetterToScroller();
  }
  doScroll(moveCnt);
*/  

  spFinish(&gxp);
  *gxp--;            /* to kill the gSPEndDisplayList spFinish adds */

  *glistp = gxp;

}


/*
 * drawobject --  Set up matricies and insert a display list
 *               of one of the objects.
 *
 * lindex is the number of the mesh to insert
 * trans and rot are for the matrices
 *
 */

void draw3DObject(dynamic_stuff *generate, Gfx **glistpParam, Gfx *dpl, vec3_t trans, vec3_t rot) {
  Gfx   *glistp;	// pointer to draw list
  u16   perspnorm;	// perspective settings
  int i;

  glistp = *glistpParam;  /* Copy to local var */
  /*
   * Set up matricies
   */
  // Set up the perspective. 
  // 750 is near clip plane
  // 7500 is far clip plane
  guPerspective(&(generate->projection), &perspnorm,
                45.0, 320.0/240.0, 50.0, 7500.0, 1.0);
  gSPPerspNormalize(glistp++, perspnorm);

  // Put rotation in dynamic data
  // (first number is rotation, second numbers are % of rotation applied to X,Y,Z axis)
  guRotate(&(generate->modeling_rotate[lastent][0]), rot[0], 1.0, 0.0, 0.0);
  guRotate(&(generate->modeling_rotate[lastent][1]), rot[1], 0.0, 1.0, 0.0);
  guRotate(&(generate->modeling_rotate[lastent][2]), rot[2], 0.0, 0.0, 1.0);
  // Put translation in dynamic data
  guTranslate(&(generate->modeling_translate[lastent]), trans[0], trans[1], trans[2]-350.0);

  // Set pointers in drawling list to right entity in dynamic data to use righ matrices
  gSPMatrix(glistp++, OS_K0_TO_PHYSICAL(&(generate->projection)),
	    G_MTX_PROJECTION|G_MTX_LOAD|G_MTX_NOPUSH);
  gSPMatrix(glistp++, OS_K0_TO_PHYSICAL(&(generate->modeling_translate[lastent])),
	    G_MTX_MODELVIEW|G_MTX_LOAD|G_MTX_NOPUSH);
  gSPMatrix(glistp++, OS_K0_TO_PHYSICAL(&(generate->modeling_rotate[lastent][0])),
	    G_MTX_MODELVIEW|G_MTX_MUL|G_MTX_NOPUSH);
  gSPMatrix(glistp++, OS_K0_TO_PHYSICAL(&(generate->modeling_rotate[lastent][2])),
	    G_MTX_MODELVIEW|G_MTX_MUL|G_MTX_NOPUSH);
  gSPMatrix(glistp++, OS_K0_TO_PHYSICAL(&(generate->modeling_rotate[lastent][1])),
	    G_MTX_MODELVIEW|G_MTX_MUL|G_MTX_NOPUSH);

  /* 
   * Set right geometry mode. We're not using smooth shading because only face normals
   * were calculated in the asc2c conversion program. (it looks better with letters)
   *
   * TODO: Could probably take it out of this function for speedup
   */
  gSPSetGeometryMode(glistp++, G_SHADE | G_ZBUFFER | G_LIGHTING | G_CULL_BACK);
//  gSPSetGeometryMode(glistp++, G_SHADE | G_ZBUFFER | G_CULL_BACK);

  /* 
   * Insert reference to display list of this mesh
   * 
   * This basically inserts a display list to display one of the letters into our main (dynamic)
   * display list. It will be affected by the translation and rotations above so it will be
   * displaced from other letters. (to move it the vertex data doesn't have to be changed)
   */
  //gSPDisplayList(glistp++, OS_K0_TO_PHYSICAL(dpl));
  gSPDisplayList(glistp++, dpl);

  *glistpParam = glistp;  /* Copy back from local var */

  // move to next entity (set of rotation and translation parms)
  lastent++;
} /* drawobject */


void strconvert(char *string, int len)
{
	unsigned long i;
	
	i=0;
	while (i<len) {
		if (string[i] == 0xf) {
			string[i] = ' ';
		} else if (string[i] == '\0') {
			string[i] = '\0';
		} else if ((string[i] >= 0x10) && (string[i] <= 0x19)) {
			string[i] = (string[i] - 0x10 + '0');
		} else {
			string[i] = (string[i] - 0x1a + 'A');
		} 
		i++;		
	}
}

void deas_strcpy(char *dest, char *orig)
{
	char *wo;
	char *wd;
	
	wo=orig;
	wd=dest;
	while (*wo != '\0') {
		*wd = *wo;
		wd++;
		wo++;
	}
}

double pow(double a, int b)
{
	int i;
	double ret;
	
	ret=1.0;
	for (i=0;i<b;i++) {
		ret = ret*a;
	}
	return ret;
}

void getnames(void)
{
	int i, k, f, j;
 	i=0;       // i is the controller where the mempack is located 
  	if (pfs[i].status == 1) {
		osPfsNumFiles(&pfs[i], &max_files, &files_used);
		for (f=0;f<files_used;f++) {
			if (osPfsFileState(&pfs[i], f, &filestate) != 0) {
				for (j=0;j<PFS_FILE_NAME_LEN;j++) {
					names[(f*20)+j] = 0;
					name[(f*20)+j] = 0;
				}
				fsize[f]=0;
				fgc[f]=0;
				fcc[f]=0;
				for (j=0;j<PFS_FILE_EXT_LEN;j++) {
					ext[(f*20)+j] = 0;
				}
			} else {
				for (j=0;j<PFS_FILE_NAME_LEN;j++) {
					names[(f*20)+j] = filestate.game_name[j];
					name[(f*20)+j] = filestate.game_name[j];
				}
				fsize[f]=filestate.file_size;
				fgc[f]=filestate.game_code;
				fcc[f]=filestate.company_code;
				for (j=0;j<PFS_FILE_EXT_LEN;j++) {
					ext[(f*20)+j] = filestate.ext_name[j];
				}
				times=0;
			}
		}
	}
	strconvert(names, 20*PFS_FILE_NAME_LEN);
}




/*
 * The work horse of this demo.
 *
 */

DemoLoop(void)
{

  dynamic_stuff *generate;
  OSTask        *gentask;
  Gfx           *glistp0;
  Gfx           *glistp;

  int           oddframe = 0;
  f32           moveCnt  = 0;

  vec3_t        Dtrans,Ntrans,Xtrans;
  vec3_t        tube1trans,tube2trans,tube3trans;

  vec3_t        Drot,Nrot,Xrot;
  vec3_t        tube1rot,tube2rot,tube3rot;
  
  int i, j, f, k, xx;
  u8 status[256];
  u8 status2[256];
  char t1[256], t2[256];
  s32 fileno;  
  
  s32 nfno;		// new file no
  
  int names_offset=0;
  int ext_offset=0;
  int active=0;
  int wait=0;
  u8 ngn[20];
  u8 nen[6];
  u16 ncc;
  u32 ngc;
  unsigned int nln;
  int ok;  
  
  fb.startx=2;
  fb.starty=1;
  fb.color=WHITE;
  fb.screen_wd=SCREEN_WD;
  fb.screen_ht=SCREEN_HT;
  fb.fb0=cfb_a;
  fb.fb1=cfb_b;

  VectorCopy(vec3_origin, Drot);
  VectorCopy(vec3_origin, Nrot);
  VectorCopy(vec3_origin, Xrot);
  VectorCopy(vec3_origin, Dtrans);
  VectorCopy(vec3_origin, Ntrans);
  VectorCopy(vec3_origin, Xtrans);

  Dtrans[1]=-35;
  Ntrans[1]=-35;
  Xtrans[1]=-35;
  
  strcpy(status, "");
  strcpy(status2, "");
  
  
  InitPrintf();
//  printf("controller data\n");
//  getprintf(1, 4, bb);
//  printf(bb);

  initControllers();
  getnames();
  
//  printf("\nthe end\n");

  osViBlack(0);
  

  while (1)
  {

/* deass section begin */  	
    if (inhelp == 1) {
    	anz=1;
    }

    if (inhelp == 0) {
	    fb.x=fb.startx+7;
	    fb.y=fb.starty;
	    s_printf(&fb,oddframe,"MempackManager for Jr");
	    fb.x=fb.startx+10;
	    fb.y=fb.starty+1;
	    s_printf(&fb,oddframe,"by deas in 1999");
	    
	    fb.x=fb.startx+2;
	    fb.y=fb.starty+12;
	    s_printf(&fb,oddframe,status);
	    
	    fb.x=fb.startx+2;
	    fb.y=fb.starty+13;
	    s_printf(&fb,oddframe,status2);
	    
	    names_offset = 0;
	    ext_offset = 0;
	    
	    for (i=0;i<files_used;i++) {
	    	fb.x=fb.startx+5;
	    	fb.y=fb.starty+3+i;
	//    	s_printf(&fb,oddframe,"%s", &names[i]);
	
	
	    	for (j=0;j<PFS_FILE_NAME_LEN;j++) {
	    		if (names[(i*20)+j] != '\0') {
				s_printf(&fb,oddframe,"%c", names[(i*20)+j]);
			}
		}
		if (fsize[i] != 0) {
			fb.x=fb.startx+30;
	    		fb.y=fb.starty+3+i;
			s_printf(&fb, oddframe, "%d", fsize[i] / 256);
		}
	    }
    } else {
	
    }
    
    	
    
    ReadController();        
    if (controllerdata[0].stick_y < -78) {
    	wait++;
    	
    	if (wait >= 10) {
    		wait=0;
    		anz=0;
	    	if (active < 10) {
	    		active=active + 1; 
	    	} else {
	    		active=0;
	    	}
	}
    }
    if (controllerdata[0].stick_y > 78) {
    	wait++;
    	
    	if (wait >= 10) {
    		wait=0;
    		anz=0;
	    	if (active > 0) {
	    		active=active - 1; 
	    	} else {
	    		active=10;
	    	}
	}
    }
    
    if (controllerdata[0].button == A_BUTTON) {
/*    	
    	if (inhelp == 0) {
    		inhelp=1;
	} else {
		inhelp=0;
	}
*/	
    }
    if (controllerdata[0].button == U_JPAD) {
        if (anz == 0) {
        	if (fsize[active] > 512) {
        		strcpy(status, "ERROR: Eeprom size <= 512!!!");
        	} else {
    			errv=osPfsReadWriteFile(&pfs[0], active, PFS_READ, 0, fsize[active], filebuffer);
    			if (errv != 0) {
	    			if (errv == PFS_ERR_BAD_DATA) {
					strcpy(status, "ERROR: bad data");
				} else {
					strcpy(status, "ERROR: unknown (write)");
				}
			} else {
				errv=osEepromLongWrite(&n_siMessageQ, 0, filebuffer, 512);
				if (errv != 0) {
					strcpy(status, "Eeprom copy failed!");
				} else {
					strcpy(status, "copied to Eeprom");
				}
			}
    		}
    		anz=1;
    	}
    }
    	
    if (controllerdata[0].button == D_JPAD) {
    	if (anz == 0) {
		errv=osEepromProbe(&n_siMessageQ);
		if (errv == 1) {
			strcpy(status, "Eeprom exist");
			errv=osEepromLongRead(&n_siMessageQ, 0, filebuffer, 512);
			if (errv == 0) {
//				sprintf(status, "%x %x %x %x %x %x %x %x", filebuffer[0], filebuffer[1], filebuffer[2], filebuffer[3], filebuffer[4], filebuffer[5], filebuffer[6], filebuffer[7]);
				nen[0] = nen[1] = nen[2] = nen[3] = 0;
				for (i=0;i<PFS_FILE_NAME_LEN;i++) {
					ngn[i] = 0;
				}
				ngn[0] = 'E'-'A'+0x1a;
				ngn[1] = 'E'-'A'+0x1a;
				ngn[2] = 'P'-'A'+0x1a;
				ngn[3] = 'R'-'A'+0x1a;
				ngn[4] = 'O'-'A'+0x1a;
				ngn[5] = 'M'-'A'+0x1a;
				nln = 512;
				errv=osPfsAllocateFile(&pfs[0], 2, 4, ngn, nen, nln, &nfno);
				xxx++;
		//		sprintf(status2, "%d", xxx);
				if (errv != 0) {
					if (errv == PFS_ERR_EXIST) {
						strcpy(status, "ERROR: File exist!");
					} else if (errv == PFS_DIR_FULL) {
						strcpy(status, "ERROR: Directory full!");
					} else {
						strcpy(status, "ERROR: unknown (allocate)");
					}
				} else {
					errv=osPfsReadWriteFile(&pfs[0], nfno, PFS_WRITE, 0, nln, filebuffer);
					if (errv != 0) {
						if (errv == PFS_ERR_BAD_DATA) {
							strcpy(status, "ERROR: bad data");
						} else {
							strcpy(status, "ERROR: unknown (write)");
						}
					} else {
					}
				}
				getnames();
			} else {
				strcpy(status, "Eeprom read error");
			}
			
		} else {
			strcpy(status, "Eeprom does not exist");
		}
        	anz = 1;
        }
    }
    
    if (controllerdata[0].button == L_CBUTTONS) {
        if (anz == 0) {
        	if (osPfsFileState(&pfs[0], active, &filestate) != 0) {
			sprintf(status, "ERROR");
		} else {
			errv=osPfsDeleteFile(&pfs[0], filestate.company_code, filestate.game_code, filestate.game_name, filestate.ext_name);
	        	if (errv == 0) {
	        		strcpy(status, "deleted...");
	        	} else {
	        		strcpy(status, "ERROR");
	        	}
		}
		osPfsChecker(&pfs[0]);
	
        	getnames();
        	anz=1;
        }
    }
    
    if (controllerdata[0].button == U_CBUTTONS) {
        if (anz == 0) {
        	j=0;
        	getprintf(0, 8, ngn);
        	if ((ngn[0] == '$') && (ngn[1] == '$') && (ngn[2] == 'F') && (ngn[3] == 'I') && (ngn[6] == '$') && (ngn[7] == '$')) {
	        	getprintf(8, 16, ngn);
	//        	sprintf(status, "%x %x %x %x %x %x %x %x", ngn[0], ngn[1], ngn[2], ngn[3], ngn[4], ngn[5], ngn[6], ngn[7]);
	        	getprintf(24, 4, nen);
	        	
	        	*t1='x';
	        	i=0;
			while (*t1 != '$') {
				getprintf(29+j, 1, t1);
				t2[i]=t1[0];
				i++;
				j++;
				
			}
			t2[i-1]='\0';
			nln=0;
			for (k=0;k<i-1;k++) {
				nln = nln + (t2[k]-'0')*pow(10, i-2-k);
			}
			
			i=0;
			*t1='x';
			while (*t1 != '$') {
				getprintf(29+j, 1, t1);
				t2[i]=t1[0];
				i++;
				j++;
				
			}
			t2[i-1]='\0';
	//		strcpy(status, t2);
			ngc=0;
			for (k=0;k<i-1;k++) {
				ngc = ngc + (t2[k]-'0')*pow(10, i-2-k);
			}
			
			i=0;
			*t1='x';
			while (*t1 != '$') {
				getprintf(29+j, 1, t1);
				t2[i]=t1[0];
				j++;
				i++;
				
			}
			t2[i-1]='\0';
			ncc=0;
			for (k=0;k<i-1;k++) {
				ncc = ncc + (t2[k]-'0')*pow(10, i-2-k);
			}
/*			
			getprintf(29+j, nln, filebuffer2);
			if (osPfsReadWriteFile(&pfs[0], active, PFS_READ, 0, fsize[active], filebuffer) == 0) {
//				sprintf(status, "%2x%2x%2x%2x - %2x%2x%2x%2x", filebuffer[2000], filebuffer[2001], filebuffer[2002], filebuffer[2003], filebuffer2[2000], filebuffer2[2001], filebuffer2[2002], filebuffer2[2003]);
//				sprintf(status, "%c%c - %c%c", filebuffer[0], filebuffer[1], filebuffer2[0], filebuffer2[1]);
				ok=-1;
				for (i=0;i<nln;i++) {
					if (filebuffer[i] != filebuffer2[i]) {
						ok=i;
						i=nln;
					} 
				}
				if (ok == -1) {
					sprintf(status, "gleich nln=%d", nln);
					osPfsFileState(&pfs[0], active, &filestate);
					if (filestate.game_code != ngc) {
						sprintf(status, "gc ungleich");
					}
					if (filestate.company_code != ncc) {
						sprintf(status, "cc ungleich");
					}
					if (filestate.file_size != nln) {
						sprintf(status, "fsize ungleich");
					}
					for (i=0;i<16;i++) {
						if (ngn[i] != filestate.game_name[i]) {
							ok=1;
							i=16;
						}
					}
					for (i=0;i<4;i++) {
						if (nen[i] != filestate.ext_name[i]) {
							ok=2;
							i=4;
						}
					}
					i=256;
					errv=osPfsGetLabel(&pfs[0], status2, &i);
					sprintf(status, "error: %d label=%s, %d", ok, status2, errv);
				} else {
					i=16336;
					sprintf(status, "ungleich bei '%d'", ok);
//					sprintf(status, "%2x%2x%2x%2x - %2x%2x%2x%2x", filebuffer[i+0], filebuffer[i+1], filebuffer[i+2], filebuffer[i+3], filebuffer2[i+0], filebuffer2[i+1], filebuffer2[i+2], filebuffer2[i+3]);
				}
			} else {
				strcpy(status, "ERROR!!!");
			}
*/


			getprintf(29+j, nln, filebuffer);
			errv=osPfsAllocateFile(&pfs[0], ncc, ngc, ngn, nen, nln, &nfno);
			xxx++;
	//		sprintf(status2, "%d", xxx);
			if (errv != 0) {
				if (errv == PFS_ERR_EXIST) {
					strcpy(status, "ERROR: File exist!");
				} else if (errv == PFS_DIR_FULL) {
					strcpy(status, "ERROR: Directory full!");
				} else {
					sprintf(status, "ERROR: unknown (allocate) %d", errv);
				}
			} else {
				errv=osPfsReadWriteFile(&pfs[0], nfno, PFS_WRITE, 0, nln, filebuffer);
				if (errv != 0) {
					if (errv == PFS_ERR_BAD_DATA) {
						strcpy(status, "ERROR: bad data");
					} else {
						strcpy(status, "ERROR: unknown (write)");
					}
				} else {
					for (i=0; i<30; i++)
      						WritePI ( PUTCHAR_BUF_START + (i*4), 0 );
				}
				osPfsChecker(&pfs[0]);
			}
			getnames();
		} else {
			strcpy(status, "no data found in buffer!");
		}
		anz=1;
        }
    }
    
    if (controllerdata[0].button == D_CBUTTONS) {
	if (anz == 0) {
          printf("$$FILE$$");
       
          for (j=0;j<PFS_FILE_NAME_LEN;j++) {
          	(void) putchar ( (int)name[(active*20)+j]) ;
	  } 
	  for (j=0;j<PFS_FILE_EXT_LEN;j++) {
	  	(void) putchar ( (int)ext[(active*20)+j]) ;
	  } 
	  printf("$");
	  printf("%d", fsize[active]);
	  printf("$");
	  printf("%d", fgc[active]);
	  printf("$");
	  printf("%d", fcc[active]);
	  printf("$");
	  if (osPfsReadWriteFile(&pfs[0], active, PFS_READ, 0, fsize[active], filebuffer) == 0) {
	  	for (j=0;j<fsize[active];j++) {
//  			printf("%2x", filebuffer[j]);
  			(void) putchar ( (int)filebuffer[j]) ;
	  	}
	  	
          }
/*          
          for (j=0;j<=32000;j++) {
          	(void) putchar ( (int)'d') ;
          }
*/          
          EndPrintf();
	  
	  anz=1;
        }
    }
    fb.x=fb.startx+3;
    fb.y=fb.starty+3+active;
    s_printf(&fb,oddframe,"o");

/* deass section end */

  lastent  = 0;
  generate = &(dynamic[(oddframe ? 0 : 1)]);
  glistp0  = &(generate->glist[0]);
  glistp   = glistp0;

  /* Tell RCP where each segment is */
  gSPSegment(glistp++, 0, 0x0);
  gSPSegment(glistp++, STATIC_SEGMENT, osVirtualToPhysical(staticSegment));

  /* Clear zbuffer */
  gSPDisplayList(glistp++, clearzbuffer_dl);

  /* Clear frame buffer */
  gDPSetColorImage(glistp++, G_IM_FMT_RGBA, G_IM_SIZ_16b, SCREEN_WD,
                   OS_K0_TO_PHYSICAL(oddframe ? cfb_a : cfb_b));
  gSPDisplayList(glistp++, clearcfb_dl);


  /* Initialize RSP and RDP state */
  gSPDisplayList(glistp++, init_dl);

  gDPSetCycleType(glistp++,G_CYC_1CYCLE);
  gDPSetRenderMode(glistp++,G_RM_AA_ZB_XLU_SURF, G_RM_AA_ZB_XLU_SURF2);
  gDPSetColorDither(glistp++, G_CD_DISABLE);
  gDPSetCombineMode(glistp++, G_CC_DECALRGBA, G_CC_DECALRGBA);
 
          gSPTexture(glistp++, 0x8000,0x8000, 0,0, G_ON);
          gDPLoadTextureBlock(glistp++, blue1, G_IM_FMT_RGBA, G_IM_SIZ_16b,
                         32,32, 0, G_TX_WRAP|G_TX_NOMIRROR, G_TX_WRAP|G_TX_NOMIRROR,
                         5,5, G_TX_NOLOD, G_TX_NOLOD);
          draw3DObject(generate, &glistp, D_dl, Dtrans, Drot);
          
          gDPLoadTextureBlock(glistp++, blue2, G_IM_FMT_RGBA, G_IM_SIZ_16b,
                         32,32, 0, G_TX_WRAP|G_TX_NOMIRROR, G_TX_WRAP|G_TX_NOMIRROR,
                         5,5, G_TX_NOLOD, G_TX_NOLOD);
          draw3DObject(generate, &glistp, N_dl, Ntrans, Nrot);

          gDPLoadTextureBlock(glistp++, blue3, G_IM_FMT_RGBA, G_IM_SIZ_16b,
                         32,32, 0, G_TX_WRAP|G_TX_NOMIRROR, G_TX_WRAP|G_TX_NOMIRROR,
                         5,5, G_TX_NOLOD, G_TX_NOLOD);
          draw3DObject(generate, &glistp, X_dl, Xtrans, Xrot);

  gDPPipeSync(glistp++);
  drawSprites(&glistp,moveCnt);
  moveCnt++;

  /* rotate DNX over z */
  Drot[2]++;Nrot[2]++;Xrot[2]++;
    
  /* rotate D and X over x backwards */
  Drot[0]-=5;Xrot[0]-=5;
    
  /* rotate N over x forwards */
  Nrot[1]+=5;

  gDPFullSync(glistp++);
  gSPEndDisplayList(glistp++);


  /* Build graphics task to execute the display list we just built */
  gentask = &(task[oddframe ? 0 : 1]);
  gentask->t.type            = M_GFXTASK;
  gentask->t.flags           = 0x0;
  gentask->t.ucode_boot      = (u64*) rspbootTextStart;
  gentask->t.ucode_boot_size = ((int)rspbootTextEnd - (int)rspbootTextStart);
  gentask->t.ucode           = (u64*) gspFast3DTextStart; /* use XBUS */
  gentask->t.ucode_data      = (u64*) gspFast3DDataStart;
  gentask->t.ucode_size      = 4096;
  gentask->t.ucode_data_size = 2048;
  gentask->t.dram_stack      = (u64*) dram_stack;
  gentask->t.dram_stack_size = 1024;
  gentask->t.output_buff     = (u64*) NULL;
  gentask->t.output_buff_size= (u64*) NULL;
  gentask->t.yield_data_ptr  = (u64*) NULL; /* Graphics only - no yielding */
  gentask->t.yield_data_size = 0x0;
  gentask->t.data_ptr        = (u64*) glistp0;
  gentask->t.data_size       = ((int) glistp - (int) glistp0);


  /* Write back dirty cache lines that need to be read by the RCP */
  osWritebackDCache(generate, sizeof(dynamic_stuff));

  /* Fire up SP task */
  osSpTaskStart(gentask);

  /* wait for SP completion */
  (void)osRecvMesg(&rspMessageQ, NULL, OS_MESG_BLOCK);
  (void)osRecvMesg(&rdpMessageQ, NULL, OS_MESG_BLOCK);

  /* setup to swap buffers */
  osViSwapBuffer(oddframe ? cfb_a : cfb_b);

  /* Make sure there isn't an old retrace in queue */
  if (MQ_IS_FULL(&retraceMessageQ))
    (void)osRecvMesg(&retraceMessageQ, NULL, OS_MESG_BLOCK);
  /* Wait for Vertical retrace to finish swap buffers */
  (void)osRecvMesg(&retraceMessageQ, NULL, OS_MESG_BLOCK);

  oddframe ^= 1;

  }

}


/*
 * The main thread, copies static segment to RAM and fires up DemoLoop
 *
 */

void mainproc(void *arg)
{
  int i;

  /* DMA static segment to ram */
  osPiStartDma(&dmaIOMessageBuf, OS_MESG_PRI_NORMAL, OS_READ,
               (u32)_staticSegmentRomStart, staticSegment,
	       (u32)_staticSegmentRomEnd - (u32)_staticSegmentRomStart,
	       &dmaMessageQ);

  /* Wait for DMA complete */
  (void)osRecvMesg(&dmaMessageQ, NULL, OS_MESG_BLOCK);


  /* Init scroller array */
//  scrollerInit();

  /* Start rocking */
  DemoLoop();

}
